/****************************************************************/
/*                                                              */
/*    NAME: SCDMixed                                            */
/*   TITLE: Macro to Diagnose Influential Subjects in           */
/*          Longitudinal Studies                                */
/* Authors: Ja-An Lin (Dept. of Biostatistics, UNC-CH)          */
/*          Allen McDowell (SAS Institute, Cary, NC)            */
/*          Grant Schneider (OSU)                               */
/* SUPPORT: Randy.Tobias@sas.com                                */
/*          schneider.393@buckeyemail.osu.edu                   */
/* UPDATE:  15 Apr 2014                                         */
/*                                                              */
/****************************************************************/
/*
SAS INSTITUTE INC. IS PROVIDING YOU WITH THE COMPUTER SOFTWARE
CODE INCLUDED WITH THIS AGREEMENT ("CODE") ON AN "AS IS" BASIS,
AND AUTHORIZES YOU TO USE THE CODE SUBJECT TO THE TERMS HEREOF.
BY USING THE CODE, YOU AGREE TO THESE TERMS.  YOUR USE OF THE
CODE IS AT YOUR OWN RISK.  SAS INSTITUTE INC. MAKES NO
REPRESENTATION OR WARRANTY, EXPRESS OR IMPLIED, INCLUDING, BUT
NOT LIMITED TO, WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE, NONINFRINGEMENT AND TITLE, WITH RESPECT TO
THE CODE.

The Code is intended to be used solely as part of a product
("Software") you currently have licensed from SAS Institute Inc.
or one of its subsidiaries or authorized agents ("SAS"). The
Code is designed to either correct an error in the Software or
to add functionality to the Software, but has not necessarily
been tested.  Accordingly, SAS makes no representation or
warranty that the Code will operate error-free.  SAS is under
no obligation to maintain or support the Code.

Neither SAS nor its licensors shall be liable to you or any
third party for any general, special, direct, indirect,
consequential, incidental or other damages whatsoever arising
out of or related to your use or inability to use the Code,
even if SAS has been advised of the possibility of such
damages.

Except as otherwise provided above, the Code is governed by
the same agreement that governs the Software.  If you do not
have an existing agreement with SAS governing the Software,
you may not use the Code.

SUMMARY

The %SCDMixed macro computes the Scaled Cook's Distance (SCD) measure
for assessing the influence of subjects in a longitudnal study.  SCD,
defined by Zhu, Ibrahim, and Cho (2012), is an extension of the familiar
Cook's Distance for guaging the effect of dropping individual
observations from a study.  SCD accounts for the fact that dropping
different subjects leads to different degrees of perturbation for the 
maximum likelihood model. For an introduction to the method and for more
information on this macro, see Schneider and Tobias (2014).

USAGE

The following example uses the Heart Rate data that was examined in
Example 10.4 of Littell et al. (2006).                                              
                                                            
   data HR;                                                     
      input patient drug$ basehr hr1 hr5 hr15 hr30 hr1h;        
      array hra{5} hr1 hr5 hr15 hr30 hr1h;                      
      do i = 1 to 5;                                            
         if (i = 1) then minute = 1/60;                         
         else if (i = 2) then minute = 5;                       
         else if (i = 3) then minute = 15;                      
         else if (i = 4) then minute = 30;                      
         else minute = 60;                                      
         time   = minute;                                       
         hours  = minute / 60;                                  
         hours1 = hours;                                        
         HR = hra{i};                                           
         output;                                                
      end;                                                      
      drop i hr1 hr5 hr15 hr30 hr1h;                            
   datalines;                                                
   201 placebo   92  76  84 88 96 84                             
   202 test      54  58  60 60 60 64                                
   203 placebo   84  86  82 84 86 82                             
   204 standard  72  72  68 68 78 72                            
   205 test      80  84  84 96 92 72                                
   206 placebo   68  72  68 68 64 62                             
   207 standard 100 104 100 92 92 68                            
   208 standard  60  60  58 56 50 56                            
   209 standard  88 104  88 88 78 84                            
   210 test      92  82  82 76 82 80                                
   211 test      88  80  84 80 80 78                                
   212 placebo  102  86  86 96 86 88                             
   214 standard  84  92 100 88 88 80                            
   215 test     104 100  96 88 92 84                                
   216 standard  92  80  72 64 68 64                            
   217 placebo   92  88  84 76 88 84                             
   218 standard  72  84  78 80 80 76                            
   219 test      72 100  92 84 88 80                                
   220 placebo   80  80  80 78 80 78                             
   221 placebo   72  68  76 72 72 68                             
   222 test      88  88  98 98 96 88                                
   223 test      88  88  96 88 88 80                                
   224 placebo   88  78  84 64 68 64                             
   232 standard  78  72  72 78 80 68                            
   ;                                                            

To use %SCDMixed, you first need to define an adjunct %SCDModel macro
with a specific structure that uses PROC MIXED with the INFLUENCE
option to fit your model. The following statements show how to do
this:

   %macro SCDModel(Datam=, Effectm=, Outpm=, nSub=);
      proc mixed data=&Datam method=ml;
         class drug hours &Effectm;
         model hr = drug basehr / noint
            influence(effect=&Effectm est)
            covb
            outpm=&Outpm;
         repeated hours /type=sp(exp)(hours) sub=&Effectm 
            r=1 to &nSub by 1;
   %mend;                                                

Finally, the actual %SCDMixed call identifies the data, subject, and
fixed effects for the modeling, as well as certain options of the
SCD analysis.

   %SCDMixed(Data           =HR, 
             Effect         =patient, 
             DepVar         =hr, 
             Fixed          =drug basehr /noint, 
             Class          =drug hours patient, 
             PlotVar        =hours, 
             PlotPercentile =100,
             TablePercentile=50,
             HighlightedSub =202 208 216 219);

The results include
   * a table of the subjects in the top 50th percentile
     of influence and
   * a plot that shows both the SCD and the pertubation
     for each subject, as well as the response trace of
     the subjects with the most influential ones
     highlighted.

REFERENCES 

Cook, R. D. (1977), "Detection of Influential Observations in
    Linear Regression," Technometrics, 19, pp. 15-18. 

Littell, R. C., Milliken, G. A., Stroup, W. W., Wolfinger, R. D.,
    and Schabenberger, O. (2006), SAS for Mixed Models, 2nd Edition,
    Cary, NC: SAS Institute Inc. 

Schneider, G. and Tobias, R. D. (2014), "A SAS Macro to Diagnose
    Influential Subjects in Longitudinal Studies," in Proceedings of
    the SAS Global Forum 2014 Conference, Cary, NC: SAS Institute Inc.
URL http://support.sas.com/resources/papers/proceedings14/1461-2014.pdf

Zhu, H., Ibrahim, J. G., and Cho, H. (2012), "Perturbation and
    Scaled Cook," Annals of Statistics, 40, pp. 785-811.
*/

%macro SCDMixed(Data=,
                Effect=,
                DepVar=,
                Fixed=,
                Class=,
                Iteration=100,
                OutTable=scd,
                method=First Order,
                PlotVar= ,
                PlotPercentile= 98 98,
                TablePercentile= 90 90,
                HighlightedSub= ,
                NoOut=0);
%local notesopt; %let notesopt = %sysfunc(getoption(notes));
options nonotes;

/************************************************************
If the user hasn't given a Plot Variable, use observed order.
************************************************************/
%let pvcheck=%length(&PlotVar);
%if &pvcheck=0 %then %do;
   %let PlotVar=wobs;
%end;


/*******************************************
Make User Specified Subjects Highlighted.
********************************************/

%let wordcheck=%length(&HighlightedSub);
%if &wordcheck=0 %then %do;
   data highlights;
   set &Data(obs=1 keep=&Effect);
   greenind=.;
   %let NoHighlights=1;
   run;
%end;
%else %do;
   %let NoHighlights=0;
   data _null_;
      call symput("numwords",countw("&HighlightedSub"));
   run;
   data highlights;
      &Effect=%scan(&HighlightedSub,1);
      greenind=1;
   run;
   %do i=2 %to &numwords;
      data tmp;
         &Effect=%scan(&HighlightedSub,&i);
         greenind=1;
      run;
      proc append base=highlights data=tmp;
      run;
   %end;
%end;

/*******************************************
Handle Percentiles for Influence Statistics
********************************************/

%let SCDPlotPercentile=%scan(&PlotPercentile,1);
%let PerturbPlotPercentile=%scan(&PlotPercentile,2);
%if %length(&PerturbPlotPercentile)=0 %then %do;
   %let PerturbPlotPercentile=&SCDPlotPercentile;
%end;
%let SCDTablePercentile=%scan(&TablePercentile,1);
%let PerturbTablePercentile=%scan(&TablePercentile,2);
%if %length(&PerturbTablePercentile)=0 %then %do;
   %let PerturbTablePercentile=&SCDTablePercentile;
%end;


proc sort data=&Data out=&Data;
  by &Effect;
run;


/******************************************************************
Convert Effect into 1,2,...,nObs in case the Data aren't contiguous
******************************************************************/
data new&Data;
  retain temporaryID 0;
  set &Data;
  by &Effect;
  if first.&Effect then temporaryID=temporaryID+1;
run;

/*******************************************************************************************
Sort the input data set by subject and generate a within-subject index of the observations. 
********************************************************************************************/
proc sort data=new&Data out=new&Data;
  by temporaryID;
run;

data new&Data;
   retain wobs 0;
   set new&Data;
   by temporaryID;
   if first.temporaryID then wobs=0;
   wobs+1;
run;

/*******************************************************************************************
Create global macro variables to store the number of subjects and
the number of observations within each subject.
********************************************************************************************/
  proc summary data=new&Data;
    var wobs;
    by temporaryID;
    output out=stats max=max;
  run;

  proc summary data=stats;
    var temporaryID;
  output out=subjects max=max;

  data _null_;
    set subjects;
    call symput('subjects',max);
  run;

  %let subjects=%eval(&subjects);
  %put There are &subjects  subjects;

  %do i = 1 %to &subjects;
    %global m&i;
    data _null_;
      set stats(where=(temporaryID=&i));
     call symput('m'||left(&i),max);
    run;
   %let m&i =%eval(&&m&i) ;
  %end;

/********************
Get the design matrix
*********************/

proc glmmod data=&data outparm=parm
   outdesign=design(drop=&DepVar) noprint;
   class &Class;
   model &DepVar = &Fixed;
run;


/*******************************************************************************************
Fit a model using PROC MIXED and save the influence statistics, the R covariance matrix, 
the estimated covariance of the betas, and the linear predictor to output data sets.
********************************************************************************************/

ods exclude all;
%SCDModel(Datam=new&Data, Effectm=temporaryID, outpm=outpm, nSub=&subjects)
ods output r=R influence=Influence covb=covb;
run;
ods select all;

data covb;
set covb(drop=Row Effect);
run;


%if &method=First Order %then %do;

/**********************************************************************
Transform Variance matrix data into the original complete matrix Cov(Y)
**********************************************************************/

proc sort data=R out=GroupNum nodupkey;
    by index;
    run;

data _Null_;
    set GroupNum end = eof;
    count+1;
    if eof then call symput('NumGroup',count);
     run;


%do i = 1 %to &NumGroup;
   data CovI&i;
       set R;
       if index = &i;
       drop index row;
       run;

   data Cov&i;
       set CovI&i;
       run;

   %DROPMISS( DSNIN = CovI&i, DSNOUT=Cov&i);

   proc datasets library=work nolist noprint;
         delete CovI&i;
      run;
      quit;
   %end;

proc iml;

   reset noprint;

   /*****************
   read design matrix
   *****************/
   use design;
     read all into X;
   close design;

   /****************************
   Create R and R_inverse matrix
   ****************************/

   %do i = 1 %to &NumGroup;
      use Cov&i;
      read all into CovM&i;
   %end;

   Rinv = INV(CovM1);
   
   %do i = 2 %to &NumGroup;
      Rinv = block(Rinv, INV(CovM&i));
   %end;

   /***************************************
   Generate Normal Sample based on R matrix
   ***************************************/
   call randseed(1); 
   mean = repeat({0},1,ncol(CovM1));
   NormalSample = RANDNORMAL( &Iteration, Mean, CovM1 );

   %do i = 2 %to &NumGroup;
      NormalSample = (NormalSample) || (RANDNORMAL( &Iteration, repeat({0},1,ncol(CovM&i)), CovM&i ));
   %end;

    /**************************
   Create Q and G tables
   ***************************/

   H = Rinv * X * GINV(X`*Rinv*X) * X` * Rinv;

   Q = Rinv - H;

   nobs = NROW(Q);

      G = Q * NormalSample`;

   CREATE QTable from Q;
   append from Q;

   CREATE GTable from G;
   append from G;

   free X R Rinv H Q G NormalSample;

   %do i = 1 %to &NumGroup;
      free CovM&i;
   %end;
run;
quit;




%do EffectID = 1 %to &NumGroup;

/**************************************
Compute an index for which observations
corrsepond to the subject
**************************************/

   data dataIDX; 
      set new&Data;
      idx = _N_;
   run;

   data dataIDX; 
     set dataIDX(where=(temporaryID = &EffectID));
     keep idx;
   run;


/***************************
Get Cook's D for the subject
***************************/

   data _Null_;
      set Influence;

      if temporaryID = &effectID
         then call symput('CookDTrue',CookD);
run;

/********************************************************
Matrix Calculation for Scaled Cooks Distance
********************************************************/

   proc iml;

      reset noprint;

      /**************************
     Read in the needed datasets
     **************************/
      use design;
      read all into X;
      close design;

      b = NCOL(X);
   
      use DataIDX;
      read all into idx;
      close dataIDX;

      use QTable;
      read all into Q;
      close QTable;

      use GTable;
      read all into G;
      close GTable;

      use Cov&EffectID;
      read all into CovM&EffectID;
      close Cov&EffectID;

      use covb;
      read all into covb;
      close covb;

     ***** Create QII and RII matrix ****;
      QII = Q[idx,idx];
      RII=INV(CovM&EffectID);
      INVQii = INV(QII);
      XI=X[idx,];

      testmat=XI`*RII*XI*covb;
      perturb=0.5*trace(testmat);

      /********************************************************
      Generate the Distribution of Scaled Cook's Distance
      ********************************************************/

      tmpG = G[,1]`;
      tmpG2 = G[,1];
      CookDistribution = tmpG[,idx] * INVQii * (RII - QII) * INVQii * tmpG2[idx,];

      %do i = 2 %to &iteration;
         tmpG = G[,&i]`;
         tmpG2 = G[,&i];
         tempresult = tmpG[,idx] * INVQii * (RII - QII) * INVQii *tmpG2[idx,];
         CookDistribution = CookDistribution // tempresult;
      %end;

      meanCook = CookDistribution[:];
      sdCook = SQRT((CookDistribution-meanCook)[##,+]/(&Iteration-1));
      ScaledCookD = &CookDTrue * b;
      ScaledCookD = (ScaledCookD - meanCook)/sdCook;
      ScaledCookDM = J(Nrow(CookDistribution),Ncol(CookDistribution),ScaledCookD);
      ScaledCookDistributionM = (CookDistribution-meanCook)/sdCook;
      IndexM = (ScaledCookDistributionM > ScaledCookDM);
      pseudoP = IndexM[+,+];
      pseudoP = pseudoP / &iteration;
      ID = &EffectID;
      ScaledCookD = ID || ScaledCookD || pseudoP || perturb;
      ColVarName = {"TemporaryID" "scd" "Pseudo Pvalue" "Perturbation"};

      CREATE TempScaledCookDTable from ScaledCookD[Colname=ColVarName] ;
      append from ScaledCookD;

      free X Rinv Q G QII RII NormalSample tempresult
         CookDistribution sdCook ScaledCookD ID ColVarName pseudoP ScaledCookDM 
       ScaledCookDistributionM IndexM INVQii;

   quit;

   %if &effectID = 1 %then %do;
      data scd;
         set TempScaledCookDTable;
      run;
   %end;


   %if &effectID > 1 %then %do;
      data scd;
         set scd TempScaledCookDTable;
      run;
   %end;

%end;

proc datasets library=work nolist noprint;
   delete groupnum Tempscaledcookdtable x _cntnts_;
run;

%do i = 1 %to &NumGroup;
   proc datasets library=work nolist noprint;
      delete Cov&i;
   run;
%end;
quit;

/*Convert from temporaryID back into original observations; */
data tmp;
set new&Data(keep= temporaryID &Effect wobs);
run;


data tmp2;
   set tmp;
run;

proc sort data=tmp nodupkey;
by &Effect;
drop wobs;
run;

data scd;
merge scd influence;
by temporaryid;
run;

data &OutTable;
merge tmp scd;
by temporaryID;
drop temporaryID effect wobs;
run;

data &OutTable;
   retain &Effect Nobs Iter PRESS CookD perturbation scd Pseudo_Pvalue;
   set &OutTable;
run;


%end;
/*End First Order Method*/

%if &method=Bootstrap %then %do;

/*************************************************************************************
Create a data set (named XBETA) that contains the design matrix along with the subject identifier
and the within-subject observation index. Append K copies together along with a
replication index. 
**************************************************************************************/

data outpm;
  set outpm(drop=StdErrPred DF Alpha Lower Upper Resid);
run;

proc surveyselect data=Outpm out=xbeta method=srs rate=100 reps=&Iteration;
run;

proc sort data=xbeta out=xbeta;
  by replicate temporaryID wobs;
run;


/***********************************************************************************************
The macro %COV generates the parametric bootstrap data set (named BOOTSTRAP) for the residuals.

   Loop through the subjects: 
     1. create a seperate covariance (_TYPE_=COV) data set for each subject
     2. generate K bootstrap samples of the residuals for each subject's covariance 
        data set using PROC SIMNORMAL
     3. transpose the bootstrap data sets
     4. add a subject identifier to each sample and generate a within-subject 
        observation index
     5. append each subject-specific bootstrap sample to a master bootstrap 
        data set
     6. rename the variable containing the bootstrap iteration number so that
        it matches the replicate number in the data set XBETA
     7. sort the master bootstrap data set so that it can be merged with XBETA
     8. delete unneeded temporary data sets

************************************************************************************************/

  /* Delete the data set BOOTSTRAP if it already exists.  */

  proc datasets nodetails;
    delete bootstrap;
  run;


  %do i = 1 %to &subjects;



   /* create a seperate covariance (_TYPE_=COV) data set for each subject */

    data r&i;
      set r(where=(index=&i));
    run;

    data r&i(type=cov drop=index row);
      set r&i(keep=index row col1-col&&m&i);
     length _NAME_ $ 10;
     length _TYPE_ $ 5;
      _TYPE_='COV';
      _NAME_=trim("col"||left(row));
    run;

   /*********From First Order approach*********/

   data dataIDX; set new&Data;
      idx = _N_;
   run;

   data dataIDX; 
      set dataIDX(where=(temporaryID = &i));
      keep idx;
   run;


   proc iml;

      use DataIDX;
        read all into idx;
      close dataIDX;

      use Design;
        read all into X;
      close Design;

      use r&i;
      read all into CovM&i;
      close r&i;

      use covb;
      read all into covbb;
      close covb;

      XI=X[idx,];
      RII=INV(CovM&i);

   testmat=XI`*RII*XI*covbb;
   perturb=0.5*trace(testmat);

   CREATE TempPertTable from perturb[Colname="Perturbation"];
   append from perturb;

%if &i = 1 %then %do;
   data PertTable;
       set TempPertTable;
       run;
%end;


%if &i > 1 %then %do;
   data PertTable;
       set PertTable TempPertTable;
       run;
%end;


%let seed= %eval(&i + &i*1000);


/* generate K bootstrap samples of the residuals for each subject's covariance data set */

proc simnormal data=r&i(type=cov) out=s&i seed=&seed numreal=&Iteration;
   var col1-col&&m&i;
run;

/* transpose the bootstrap data sets */

proc transpose data=s&i out=s&i;
   by rnum;
run;

/* add a subject identifier to each sample and generate a within-subject observation index */

data s&i(drop=_NAME_);
    set s&i;
    wobs=abs(strip(compress(_NAME_,'Col')));
   temporaryID=&i;
run;

   /* append each subject-specific bootstrap sample to the master bootstrap data set */

proc append base=bootstrap data=s&i;
run;

%end;


  /* rename the variable containing the bootstrap iteration number so that
     it matches the replicate number in the data set XBETA               */

  data bootstrap;
    set bootstrap;
    rename rnum=replicate col1=residual;
  run;


  /*********************************************************************
  sort the master bootstrap data set so that it can be merged with XBETA
  *********************************************************************/
  proc sort data=bootstrap out=bootstrap;
    by replicate temporaryID wobs;
  run; 

  /* clean up temporary data set */
  proc datasets nodetails;
    delete r1-r&subjects s1-s&subjects stats;
  run;




/**********************************************************
Merge the data sets BOOTSTRAP and XBETA.
Generate the response variable for the bootstrap samples.
**********************************************************/

 data bootstrap(drop=residual pred);
    merge bootstrap xbeta;
    by replicate temporaryID wobs;
    &DepVar= pred + residual;
  run;


/*************************************************************************
Fit the mixed model for each bootstrap sample using BY-group processing.
Save the influence statistics in a data set.
*************************************************************************/

ods select none;
%SCDModel(Datam=bootstrap, Effectm=temporaryID, Outpm=deletethis, nSub=&subjects);
   by replicate;
   ods output influence=bsInfluence;
run;
ods select all;

proc datasets nodetails;
   delete deletethis xbeta;
run;

   
/************************************************************
Compute the means and standard deviations of Cook's distance
using the K bootstrap samples. Merge the resulting output
data set with the original influence statistic data set and
compute the scaled Cook's distance.
************************************************************/

proc sort data=bsinfluence out=bsinfluence;
  by temporaryID;
run;

proc summary data=bsinfluence mean std;
  var cookd;
  by temporaryID;
  output out=scd(drop=_TYPE_ _FREQ_) mean=mean std=std;
run;

%do i=1 %to &subjects;

data tmppseudo;
   set bsinfluence(where=(temporaryID=&i));
run;
data tmpinfluence;
   set influence(where=(temporaryID=&i));
run;

proc iml;
use tmppseudo;
   read all var {CookD} into pseudocalc;
close tmppseudo;

use tmpinfluence;
   read all var {CookD} into pseudoval;
close tmpinfluence;

IndexM = (pseudocalc >= pseudoval);
      pseudoP = IndexM[+,+];
      pseudoP = pseudoP / &Iteration;
     pseudoOut= &i || pseudoP;

CREATE PseudoTable from pseudoOut[Colname={"TemporaryID" "Pseudo PValue"}] ;
      append from PseudoOut;
quit;

%if &i = 1 %then %do;
      data pseudoPValue;
         set PseudoTable;
      run;
%end;

%if &i > 1 %then %do;
      data pseudoPValue;
         set pseudoPValue PseudoTable;
      run;
%end;
%end;

data scd;
merge scd pseudoPValue;
by temporaryID;
label Pseudo_P_Value="Pseudo Pvalue";
run;

data scd;
  merge scd influence;
  by temporaryID;
  scd=(cookd-mean)/std;
  label scd="Scaled Cook's Distance";
run;








/*Convert from temporaryID back into original observations; */
data tmp;
set new&Data(keep= temporaryID &Effect wobs);
run;

data tmp2;
   set tmp;
run;

proc sort data=tmp nodupkey;
by &Effect;
drop wobs;
run;

data &OutTable;
merge tmp scd;
by temporaryID;
drop temporaryID mean std effect;
run;

data &OutTable;
merge &OutTable perttable;
run;

data &OutTable;
   retain &Effect Nobs Iter PRESS CookD perturbation scd Pseudo_P_Value;
   set &OutTable;
run;
proc datasets nodetails;
   delete    bootstrap;
run;
quit;
%end;
/*End Bootstrap Method*/

%if &NoOut=0 %then %do;

/***********User Output***********/
%if &PerturbPlotPercentile=&PerturbTablePercentile %then %do;
proc univariate data=&OutTable noprint;
   var perturbation;
   output out=pctlPerturb pctlpre=P_ PCTLPTS=&PerturbPlotPercentile;
run;
%end;
%else %do;
proc univariate data=&OutTable noprint;
   var perturbation;
   output out=pctlPerturb pctlpre=P_ PCTLPTS=&PerturbPlotPercentile &PerturbTablePercentile;
run;
%end;

%if &SCDPlotPercentile=&SCDTablePercentile %then %do;
proc univariate data=&OutTable noprint;
   var scd;
   output out=pctlSCD pctlpre=P_ PCTLPTS=&SCDPlotPercentile;
run;
%end;
%else %do;
proc univariate data=&OutTable noprint;
   var scd;
   output out=pctlSCD pctlpre=P_ PCTLPTS=&SCDPlotPercentile &SCDTablePercentile;
run;
%end;


data _null_;
  set pctlPerturb;
  call symput("PerturbPlotCutoff",P_&PerturbPlotPercentile);
  call symput("PerturbTableCutoff",P_&PerturbTablePercentile);
run;


data _null_;
   set pctlSCD;
   call symput("SCDPlotCutoff",P_&SCDPlotPercentile);
   call symput("SCDTableCutoff",P_&SCDTablePercentile);
run;

data &OutTable;
   merge &OutTable highlights; 
   by &Effect;
run;

data tmpout;
   merge &OutTable Influence;
   if perturbation < &PerturbPlotCutoff then do;
      scdred=.;
      pertred=.;
   end;
   if perturbation >= &PerturbPlotCutoff then do; 
      pertlabel=&Effect;
      scdlabel=&Effect;
      scdred=scd;
      pertred=perturbation;
   end;

   if scd < &SCDPlotCutoff then do;
      scdblue=.;
      pertblue=.;
   end;
   if scd >= &SCDPlotCutoff then do; 
      scdlabel=&Effect;
      pertlabel=&Effect;
      scdblue=scd;
      pertblue=perturbation;
   end;
   if greenind=1 then do;
         scdlabel=&Effect;
      pertlabel=&Effect;
      scdgreen=scd;
      pertgreen=perturbation;
   end;
run;


data scdinfluencers;
   set tmpout(where=(scdblue^=.));
   keep &Effect _x1blueind;
   _x1blueind=1;
run;
data pertinfluencers;
   set tmpout(where=(pertred^=.));
   keep &Effect _x1redind;
   _x1redind=1;
run;
data userselected;
   set tmpout(where=(scdgreen^=.));
   keep &Effect _x1greenind;
   _x1greenind=1;
run;

data &Data;
   merge &Data tmp2(keep=&Effect wobs);
   by &Effect;
run;

data tmp&Data;
   merge &Data scdinfluencers;
   by &Effect;
run;
data tmp&Data;
   merge tmp&Data pertinfluencers;
   by &Effect;
run;

data tmp&Data;
   merge tmp&Data userselected;
   by &Effect;
run;

data outfile;
   set tmpout tmp&Data;
run;

data outfile;
   set outfile;
   if _x1blueind=. then _x1blue=.;
   if _x1blueind=1  then _x1blue=&PlotVar;
   if _x1redind=. then _x1red=.;
   if _x1redind=1  then _x1red=&PlotVar;
   if _x1greenind=. then _x1green=.;
   if _x1greenind=1  then _x1green=&PlotVar;
run;


%let tmpPert=%eval(100-&PerturbPlotPercentile);
%let tmpSCD=%eval(100-&SCDPlotPercentile);


%let pertsym=" ";
%let scdsym=" ";

%let title2="Highlighted: ";
%if &tmpPert>0 %then %do; 
   %let tmptitle="top &tmpPert% perturbation ";
   %let title2 =%sysfunc(catx(%str( ), &title2, &tmptitle ));
   %let pertsym=" + ";   
%end;
%if &tmpSCD>0 %then %do; 
   %let tmptitle="top &tmpSCD% SCD ";
   %let title2 =%sysfunc(catx(%str( ), &title2, &pertsym));
   %let title2 =%sysfunc(catx(%str( ), &title2, &tmptitle ));
   %let scdsym=" + ";
%end;
%if &NoHighlights^=1 %then %do; 
   %let tmptitle="&Effect=&HighlightedSub ";
   %let title2 =%sysfunc(catx(%str( ), &title2, &scdsym));
   %let title2 =%sysfunc(catx(%str( ), &title2, &tmptitle ));
%end;


proc template;
   define statgraph scdplots;
      begingraph;
         entrytitle "Influence Diagnostics for &Data" ;
         entrytitle &title2
                        / textattrs=(size=8pt) ;
         layout gridded / rows=2 columns=2 rowgutter=20 columngutter=20;

            layout overlay /xaxisopts=(label="Scaled Cook's Distance ") yaxisopts=(label="Perturbation");
            scatterplot x=scd y=Perturbation / datatransparency=0.6 markerattrs=GraphData1;
         %if &SCDPlotPercentile<100 %then %do;
            scatterplot x=scdblue y=Perturbation / datalabel=scdlabel datatransparency=0.4 markerattrs=GraphData2 datalabelattrs=GraphData2;
         %end;
         %if &PerturbPlotPercentile<100 %then %do;
            scatterplot x=scdred y=Perturbation / datalabel=scdlabel datatransparency=0.4 markerattrs=GraphData3 datalabelattrs=GraphData3;
         %end;
         %if &NoHighlights^=1 %then %do;
            scatterplot x=scdgreen y=Perturbation / datalabel=scdlabel datatransparency=0.2 markerattrs=GraphData4 datalabelattrs=GraphData4;
         %end;
            endlayout;
                        
         
            layout overlay /xaxisopts=(label="Cluster Size") yaxisopts=(label="Perturbation ");
            scatterplot x=Nobs y=Perturbation/ datatransparency=0.6 markerattrs=GraphData1;
         %if &SCDPlotPercentile<100 %then %do;
            scatterplot x=Nobs y=pertblue / datalabel=pertlabel datatransparency=0.4 markerattrs=GraphData2 datalabelattrs=GraphData2;
         %end;
         %if &PerturbPlotPercentile<100 %then %do;
            scatterplot x=Nobs y=pertred / datalabel=pertlabel datatransparency=0.4 markerattrs=GraphData3 datalabelattrs=GraphData3;
         %end;
         %if &NoHighlights^=1 %then %do;
            scatterplot x=Nobs y=pertgreen / datalabel=scdlabel datatransparency=0.2 markerattrs=GraphData4 datalabelattrs=GraphData4;
         %end;
            endlayout;

            layout overlay /xaxisopts=(label="Scaled Cook's Distance") yaxisopts=(label="Cook's Distance");
            scatterplot x=scd y=cookd / datatransparency=0.6 markerattrs=GraphData1;
         %if &SCDPlotPercentile<100 %then %do;
            scatterplot x=scdblue y=cookd / datalabel=scdlabel datatransparency=0.4 markerattrs=GraphData2 datalabelattrs=GraphData2;
         %end;
         %if &PerturbPlotPercentile<100 %then %do;
            scatterplot x=scdred y=cookd /  datalabel=scdlabel datatransparency=0.4 markerattrs=GraphData3 datalabelattrs=GraphData3;
         %end;
         %if &NoHighlights^=1 %then %do;
            scatterplot x=scdgreen y=cookd /  datalabel=scdlabel datatransparency=0.2 markerattrs=GraphData4 datalabelattrs=GraphData4;
         %end;
            endlayout;
         
         %if &PlotVar=wobs %then %do;    
            layout overlay /xaxisopts=(label="Observed Order") yaxisopts=(label="&DepVar");
         %end;
         %else %do;
            layout overlay /xaxisopts=(label="&PlotVar") yaxisopts=(label="&DepVar");
         %end;

            seriesplot y=&DepVar x=&PlotVar / group=&Effect datatransparency=0.9 lineattrs=GraphData1;
         %if &SCDPlotPercentile<100 %then %do;
            seriesplot y=&DepVar x=_x1blue / group=&Effect datatransparency=0.4 curvelabel=&Effect 
                     lineattrs=GraphData2;
         %end;
         %if &PerturbPlotPercentile<100 %then %do;
            seriesplot y=&DepVar x=_x1red / group=&Effect datatransparency=0.4 curvelabel=&Effect 
                     lineattrs=GraphData3;
         %end;
         %if &NoHighlights^=1 %then %do;
            seriesplot y=&DepVar x=_x1green / group=&Effect datatransparency=0.2 curvelabel=&Effect 
                     lineattrs=GraphData4;            
         %end;
            endlayout;
         endlayout;
      endgraph;
   end;
run;


proc sgrender data=outfile template=scdplots;
run;



data tableout;
set &Outtable(drop=greenind);
run;

data tableout;
   if (&SCDTablePercentile=100) then do;
      set tableout(where=(perturbation>=&PerturbTableCutoff));
   end;
   if (&PerturbTablePercentile=100) then do;
      set tableout(where=(scd>=&SCDTableCutoff));
   end;
   else do;
   set tableout(where=((scd>=&SCDTableCutoff) or (perturbation>=&PerturbTableCutoff)));
   end;
run;

data tableout;
   set tableout(keep=&Effect Nobs CookD perturbation scd Pseudo_Pvalue);
run;


proc print data=tableout noobs;
run;
%end;
proc datasets noprint nolist nowarn;
   delete    bsinfluence covb dataidx design influence new&data 
         outfile outpm parm pctlperturb pctlscd pertinfluencers perttable 
         R scd scdinfluencers subjects tempperttable tmp tmphr tmpout tmp&Data
         Gtable Qtable stats;
run;
quit;



options &notesopt;

%mend SCDMixed;


/*************************************************************************
The macro DROPMISS is created in the following document :
"Dropping Automatically Variables with Only Missing Values" by
Selvaratnam Sridharma, Census Bureau, Washington, D.C.

URL:
http://support.sas.com/resources/papers/proceedings10/048-2010.pdf
************************************************************************/



%macro DROPMISS( DSNIN=, DSNOUT=  , NODROP=) ; 

*options nomprint noSYMBOLGEN MLOGIC;

%local I ; 
%if "&DSNIN" = "&DSNOUT"
     %then %do ; 
%put /------------------------------------------------\ ; 
%put | ERROR from DROPMISS:                           | ; 
%put | Input Dataset has same name as Output Dataset. | ; 
%put | Execution terminating forthwith.               | ; 
%put \------------------------------------------------/ ; 
%goto L9999 ; 
    %end ; 


/*###################################################################*/
/* begin executable code 
/*####################################################################/ 
/*===================================================================*/
/* Create dataset of variable names that have only missing values 
/* exclude from the computation all names in &NODROP 
/*===================================================================*/;


proc contents data=&DSNIN( drop=&NODROP ) memtype=data noprint out=_cntnts_( keep= name type ) noprint; 
   run ; 


%let N_CHAR = 0 ; 
%let N_NUM  = 0 ; 

data _null_ ; 
       set _cntnts_ end=lastobs nobs=nobs ; 
       if nobs = 0 then stop ; 
       n_char + ( type = 2 ) ; 
       n_num  + ( type = 1 ) ; 
       /* create macro vars containing final # of char, numeric variables */
       if lastobs 
       then do ; 
          call symput( 'N_CHAR', left( put( n_char, 5. ))) ; 
          call symput( 'N_NUM' , left( put( n_num , 5. ))) ; 
       end ; 
    run ; 



/*===================================================================*/
/* if there are no variables in dataset, stop further processing 
/*===================================================================*/


%if %eval( &N_NUM + &N_CHAR ) = 0
%then %do ; 
%put /----------------------------------\ ; 
%put | ERROR from DROPMISS:              | ; 
%put | No variables in dataset.         | ; 
%put | Execution terminating forthwith. | ; 
%put \----------------------------------/ ; 
%goto L9999 ; 
%end ; 


/*===================================================================*/
/* put global macro names into global symbol table for later retrieval 
/*===================================================================*/


%LET NUM0 =0; 
%LET CHAR0 = 0; 
%IF &N_NUM >0 %THEN %DO; 
    %do I = 1 %to &N_NUM ; 
       %global NUM&I  ; 
    %end ; 
%END; 



%if &N_CHAR > 0 %THEN %DO; 
    %do I = 1 %to &N_CHAR ; 
       %global CHAR&I  ; 
    %end ; 
%END; 


/*===================================================================*/
/* create macro vars containing variable names 
/* efficiency note: could compute n_char, n_num here, but must declare macro names to be global b4 stuffing them 
/* 
/*===================================================================*/

proc sql noprint noprint; 
       %if &N_CHAR > 0 %then %str( select name into :CHAR1 - :CHAR&N_CHAR from _cntnts_ where type = 2 ; ) ; 
       %if &N_NUM  > 0 %then %str( select name into :NUM1  - :NUM&N_NUM   from _cntnts_ where type = 1  ; ) ; 
    quit ; 
    

/*===================================================================*/
/* Determine the variables that are missing SAS Global Forum 2010 Coders' Corner5
/* 
/*===================================================================*/

%IF &N_CHAR > 1 %THEN %DO; 
      %let N_CHAR_1 = %EVAL(&N_CHAR - 1); 
      %END; 


Proc sql noprint; 
      select %do I= 1 %to &N_NUM; max (&&NUM&I) , %end; %IF &N_CHAR > 1 %THEN %DO; 
              %do I= 1 %to &N_CHAR_1; max(&&CHAR&I), %END; %end; MAX(&&CHAR&N_CHAR) into 
              %do I= 1 %to &N_NUM; :NUMMAX&I , %END; %IF &N_CHAR > 1 %THEN %DO;                       
              %do I= 1 %to &N_CHAR_1; :CHARMAX&I,%END; %END; :CHARMAX&N_CHAR 
              from &DSNIN; 
    quit; 
  
/*===================================================================*/
/* initialize DROP_NUM, DROP_CHAR global macro vars 
/*===================================================================*/
%let DROP_NUM      =  ; 
%let DROP_CHAR     =  ; 
     
%if &N_NUM > 0 %THEN %DO; 
   DATA _NULL_; 
    
     %do I = 1 %to &N_NUM ; 
         %IF &&NUMMAX&I =. %THEN %DO; 
        %let DROP_NUM = &DROP_NUM %qtrim( &&NUM&I ) ; 
       %END; 
%end ; 
RUN; 
%END; 

%IF &N_CHAR > 0 %THEN %DO; 
   DATA _NULL_; 
     %do I = 1 %to &N_CHAR ; 
         %IF "%qtrim(&&CHARMAX&I)" eq ""  %THEN %DO; 
          %let DROP_CHAR     = &DROP_CHAR %qtrim( &&CHAR&I )  ; 
     %END;  
    %end ; 
   RUN; 
%END;  
   


/*===================================================================*/
/* Create output dataset 
/*===================================================================*/
data &DSNOUT ; 
   
       %if &DROP_CHAR ^=  %then %str(DROP &DROP_CHAR ;     ) ; /* drop char variables that have only missing values      */
       %if  &DROP_NUM ^=   %then %str(DROP  &DROP_NUM ;      ) ; /* drop num variables that have only missing values  */
       set &DSNIN ; 
/*
      %if &DROP_CHAR ^=  or &DROP_NUM ^=   %then %do; 
            %put /----------------------------------\ ; 
         %put | Variables dropped are &DROP_CHAR &DROP_NUM | ;            
         %put \----------------------------------/ ; 
            %end; 
        %if &DROP_CHAR =  and &DROP_NUM =   %then %do; 
            %put /----------------------------------\ ; 
            %put  |     No variables are dropped  |;            
         %put \----------------------------------/ ; 
            %end; 
*/
      run ; 
%L9999: 
%mend DROPMISS ;
