/****************************************************************/
/* Example 1 from Schneider and Tobias (2014).                  */
/****************************************************************/

/*
This is an example of using the %SCDMixed macro for performing
an influence analysis using the Scaled Cook's Distance meaure of
Zhu, Ibrahim, and Cho (2012).  Data are simulated that contain
both noisy outliers and shifted outliers, as well as exhibiting
various kinds of unbalance.  You should run this sample code from
the directory that contains the SCDMixed.sas file.

REFERENCES 

Schneider, G. and Tobias, R. D. (2014), "A SAS Macro to Diagnose
    Influential Subjects in Longitudinal Studies," in Proceedings of
    the SAS Global Forum 2014 Conference, Cary, NC: SAS Institute Inc.
URL http://support.sas.com/resources/papers/proceedings14/1461-2014.pdf

Zhu, H., Ibrahim, J. G., and Cho, H. (2012), "Perturbation and
    Scaled Cook's Distance," Annals of Statistics, 40, pp. 785-811.
*/

%include "SCDMixed.sas";
ods graphics on / reset imagename='SimulatedData';

/*
/ The following SAS statements simulate 50 subjects with an
/ autoregressive covariance model at 200 time points per subject. To
/ account for the fact that observations on a subject are often
/ unequally spaced throughout time, keep a subset of 24 observations
/ for each subject.
/---------------------------------------------------------------------*/
data AR1;
   retain index 0 subject 1;
   do j=1 to 50;
      ytm1=rannor(12);
      do i=1 to 201;
         et=rannor(1);
         yt=0.7*ytm1+et;
         output;
         ytm1=yt;
         index+1;
      end;
      index=0;
      subject+1;
   end;
   keep yt index subject;
run;

data AR1;
   set AR1(where=(index in (0, 5, 10, 15, 20, 25, 35, 45,
                            55, 65, 75, 100, 120, 150, 155, 160, 
                            165, 170, 175, 180, 185, 190, 195, 200)));
run;

/*
/ First, introduce noisy outliers to the data by scaling subjects 9,
/ 16, 25, 36, and 49 by a factor of 5.
/---------------------------------------------------------------------*/
data ScaledAR1;
   set AR1;
   if subject in (9,16,25,36,49) then yt=5*yt;
run;

/*
/ Next, introduce shifted outliers to the data by adding 5 to
/ measurements from subjects 9, 16, 25, 36, and 49. 
/---------------------------------------------------------------------*/
data ShiftedAR1;
   set AR1;
   if subject in (9,16,25,36,49) then yt=5+yt;
run;

/*
/ The perturbation index computed by the %SCDMixed macro is most useful
/ when subjects are not observed at the same points, and thus,
/ perturbation varies from subject to subject. To investigate this
/ situation, uniformly sample a "dropout" time on the observations for
/ each subject. 
/---------------------------------------------------------------------*/
data Cutoffs;
   call streaminit(1);
   do subject = 1 to 50;
      u = rand("Uniform");
      cutoff = ceil(200*u);
      output;
    end;
run;

/*
/ The following dataset will be used to explore perturbation when there
/ are no outliers. 
/---------------------------------------------------------------------*/
data Longitudinal;
   merge Cutoffs(keep=subject cutoff) AR1;
   by subject;
   if index>cutoff then delete;
   drop cutoff;
run;

/*
/ For the final simulated dataset, introduce outliers into the
/ longitudinal data with dropouts. As before, scale subjects 9, 16, 25,
/ 36, and 49 by a factor of 5.
/---------------------------------------------------------------------*/
data ScaledLongitudinal;
   merge Cutoffs(keep=subject cutoff) ScaledAR1;
   by subject;
   if index>cutoff then delete;
   drop cutoff;
run;

/*
/ For all analyses of this simulated data, the following %SCDModel macro\
/ defines the mixed model to be fit. It is based on the true model used
/ to generate the data. 
/---------------------------------------------------------------------*/
%macro SCDModel(Datam=, Effectm=, Outpm=, nSub=);
   proc mixed data=&Datam method=ml;
      class &Effectm;
      model yt = index /
         influence(effect=&Effectm est)                  /*Required*/
         covb                                            /*Required*/
         outpm=&Outpm;                                   /*Required*/
      repeated /type=sp(exp)(index) sub=&Effectm 
         r=1 to &nSub by 1;                              /*Required*/
%mend SCDModel; 




/*Calls to SCDMixed Macro*/

%SCDMixed(Data=AR1, 
          Effect=subject, 
          DepVar=yt, 
          Fixed=index, 
          Class=subject, 
          PlotVar=index,
          PlotPercentile=90 100,
          TablePercentile=80 100);  

%SCDMixed(Data=ScaledAR1, 
          Effect=subject, 
          DepVar=yt, 
          Fixed=index, 
          Class=subject, 
          method=First Order, 
          PlotVar=index, 
          PlotPercentile=90 100,
          TablePercentile=80 100);

%SCDMixed(Data=ScaledAR1, 
          Effect=subject, 
          DepVar=yt, 
          Fixed=index, 
          Class=subject, 
          method=First Order, 
          PlotVar=index, 
          PlotPercentile=100,
          TablePercentile=80 100,
          HighlightedSub=49);

%SCDMixed(Data=ShiftedAR1, 
          Effect=subject, 
          DepVar=yt, 
          Fixed=index, 
          Class=subject, 
          method=First Order, 
          PlotVar=index, 
          PlotPercentile=90 100,
          TablePercentile=80);

%SCDMixed(Data=Longitudinal, 
          Effect=subject, 
          DepVar=yt, 
          Fixed=index, 
          Class=subject, 
          method=First Order, 
          PlotVar=index, 
          PlotPercentile=100,
          TablePercentile=90,
          HighlightedSub=16 28 40);

%SCDMixed(Data=ScaledLongitudinal, 
          Effect=subject, 
          DepVar=yt, 
          Fixed=index, 
          Class=subject, 
          method=First Order, 
          PlotVar=index, 
          PlotPercentile=90);
